/*
-----------------------------------------------------------------------------
This source file is part of the GameTools Project
http://www.gametools.org

Author: Martin Szydlowski
-----------------------------------------------------------------------------
*/

#include <OgreCamera.h>
#include <OgreSubEntity.h>
#include "OgreKdTreeSceneManager.h"
#include "OgreKdTreeSceneNode.h"
#include "OgreKdTree.h"
#include "OgreVisibilityOptionsManager.h"

#include <OgreLogManager.h>
#include <OgreStringConverter.h>

namespace Ogre
{
	KdTreeSceneManager::KdTreeSceneManager(const String& name, GtpVisibility::VisibilityManager *vm):
		SceneManager(name), 
		mVisibilityManager(vm), 
		mKdTree(0), 
		mMaxDepth(KDTREE_MAX_DEPTH),
		mShowBoxes(false),
#ifdef KDTREE_DEBUG
		mHighlighLevel(0),
		mShowAllBoxes(false),
#endif
		mBuildMethod(KdTree::KDBM_PRIORITYQUEUE),
		mRenderMethod(KdTree::KDRM_RECURSE),
		mLeavePassesInQueue(0),
		mCurrentEntityId(1)
	{
		// Replace root node with my node
		OGRE_DELETE(mSceneRoot);

		mSceneRoot = new KdTreeSceneNode(this, "root node");
		mSceneRoot->_notifyRootNode();

		mHierarchyInterface = new KdTreeHierarchyInterface(this, mDestRenderSystem);
	}

	KdTreeSceneManager::~KdTreeSceneManager(void)
	{
		delete mKdTree;
	}

	const String& KdTreeSceneManager::getTypeName(void) const
	{
		return KdTreeSceneManagerFactory::FACTORY_TYPE_NAME;
	}

	void KdTreeSceneManager::setShowBoxes(bool showboxes)
	{
		mShowBoxes = showboxes;
	}

	bool KdTreeSceneManager::getShowBoxes(void) const
	{
		return mShowBoxes;
	}

	bool KdTreeSceneManager::setOption(const String& strKey, const void* pValue)
	{
		// change max depth of the kdtree
		// rebuild the tree if already exists
		if (strKey == "KdTreeMaxDepth")
		{
			int maxdepth = *static_cast<const int *>(pValue);

			// no negative depth, plz!
			if (maxdepth < 0)
			{
				return false;
			}
			else
			{
				mMaxDepth = maxdepth;
#ifdef KDTREE_DEBUG
				if (mHighlighLevel > mMaxDepth)
					mHighlighLevel = mMaxDepth;
#endif
				return true;
			}
			return true;
		}
		else if (strKey == "KT")
		{
			Real kt = *static_cast<const Real *>(pValue);
			if (kt > 0)
			{
				PlaneEvent::KT = kt;
				return true;
			}
			else
			{
				return false;
			}
		}
		else if (strKey == "KI")
		{
			Real ki = *static_cast<const Real *>(pValue);
			if (ki > 0)
			{
				PlaneEvent::KI = ki;
				return true;
			}
			else
			{
				return false;
			}
		}
		else if (strKey == "RebuildKdTree")
		{
			OGRE_DELETE(mKdTree);
			mKdTree = new KdTree(mMaxDepth);
			mKdTree->setBuildMethod(mBuildMethod);
			mKdTree->build(static_cast<KdTreeSceneNode *>(mSceneRoot));
			return true;
		}
#ifdef KDTREE_DEBUG
		else if (strKey == "HighlightLevel")
		{
			int hl = *static_cast<const int *>(pValue);
			if (hl >= 0 && hl <= mMaxDepth)
			{
				mHighlighLevel = hl;
				return true;
			}
			else
			{
				return false;
			}
		}
		else if (strKey == "ShowAllBoxes")
		{
			bool sa = *static_cast<const bool *>(pValue);
			mShowAllBoxes = sa;
			return true;
		}
#endif
		else if (strKey == "BuildMethod")
		{
			String bm = *static_cast<const String *>(pValue);
			if (bm == "PriorityQueue")
			{
				mBuildMethod = KdTree::KDBM_PRIORITYQUEUE;
				return true;
			}
			else if (bm == "Recursive")
			{
				mBuildMethod = KdTree::KDBM_RECURSIVE;
				return true;
			}
			else
			{
				return false;
			}
		}
		else if (strKey == "RenderMethod")
		{
			String rm = *static_cast<const String *>(pValue);
			if (rm == "Recursive")
			{
				mRenderMethod = KdTree::KDRM_RECURSE;
				return true;
			}
			else if (rm == "Stack")
			{
				mRenderMethod = KdTree::KDRM_STACK;
				return true;
			}
			else if (rm == "StopAndWait")
			{
				mRenderMethod = KdTree::KDRM_SAW;
				return true;
			}
			else if (rm == "CHC")
			{
				mRenderMethod = KdTree::KDRM_CHC;
				return true;
			}
			else
			{
				return false;
			}
		}
		else if (strKey == "ShowKdTree")
		{
			bool sk = *static_cast<const bool *>(pValue);
			mShowBoxes = sk;
			return true;
		}

		return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface)
			.setOption(strKey, pValue) || SceneManager::setOption(strKey, pValue);
	}

	bool KdTreeSceneManager::getOption(const String& strKey, void* pDestValue)
	{
		if (strKey == "KdTreeMaxDepth")
		{
			*static_cast<int *>(pDestValue) = mMaxDepth;
			return true;
		}
		else if (strKey == "KT")
		{
			*static_cast<Real *>(pDestValue) = PlaneEvent::KT;
			return true;
		}
		else if (strKey == "KI")
		{
			*static_cast<Real *>(pDestValue) = PlaneEvent::KI;
			return true;
		}
#ifdef KDTREE_DEBUG
		else if (strKey == "HighlightLevel")
		{
			*static_cast<int *>(pDestValue) = mHighlighLevel;
			return true;
		}
		else if (strKey == "ShowAllBoxes")
		{
			*static_cast<bool *>(pDestValue) = mShowAllBoxes;
			return true;
		}
#endif
		else if (strKey == "BuildMethod")
		{
			if (mBuildMethod == KdTree::KDBM_PRIORITYQUEUE)
			{
				*static_cast<String *>(pDestValue) = "PriorityQueue";
			}
			else if (mBuildMethod == KdTree::KDBM_RECURSIVE)
			{
				*static_cast<String *>(pDestValue) = "Recursive";
			}
			return true;
		}
		else if (strKey == "RenderMethod")
		{
			if (mRenderMethod == KdTree::KDRM_RECURSE)
			{
				*static_cast<String *>(pDestValue) = "Recursive";
			}
			else if (mRenderMethod == KdTree::KDRM_STACK)
			{
				*static_cast<String *>(pDestValue) = "Stack";
			}
			else if (mRenderMethod == KdTree::KDRM_SAW)
			{
				*static_cast<String *>(pDestValue) = "StopAndWait";
			}
			else if (mRenderMethod == KdTree::KDRM_CHC)
			{
				*static_cast<String *>(pDestValue) = "CHC";
			}
			return true;
		}
		else if (strKey == "TreeBox")
		{
			if (mKdTree)
				*static_cast<AxisAlignedBox *>(pDestValue) = mKdTree->getBox();
			else
				*static_cast<AxisAlignedBox *>(pDestValue) = AxisAlignedBox();
		}

		return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface)
			.getOption(strKey, pDestValue) || SceneManager::getOption(strKey, pDestValue);
	}

	bool KdTreeSceneManager::getOptionKeys(StringVector &refKeys)
	{
		refKeys.push_back("BuildMethod");
		refKeys.push_back("KI");
		refKeys.push_back("KT");
		refKeys.push_back("KdTreeMaxDepth");
		refKeys.push_back("RebuildKdTree");
		refKeys.push_back("RenderMethod");
		refKeys.push_back("ShowKdTree");
		refKeys.push_back("TreeBox");
#ifdef KDTREE_DEBUG
		refKeys.push_back("HighlightLevel");
		refKeys.push_back("ShowAllBoxes");
#endif
		return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface)
			.getOptionKeys(refKeys);
	}

	bool KdTreeSceneManager::getOptionValues(const String & key, StringVector &refValueList)
	{
		return SceneManager::getOptionValues(key, refValueList);
	}

	void KdTreeSceneManager::setVisibilityManager(GtpVisibility::VisibilityManager *visManager)
	{
		mVisibilityManager = visManager;
	}

	GtpVisibility::VisibilityManager * KdTreeSceneManager::getVisibilityManager()
	{
		return mVisibilityManager;
	}

	GtpVisibility::VisibilityManager * KdTreeSceneManager::GetVisibilityManager()
	{
		return mVisibilityManager;
	}

	KdTreeHierarchyInterface * KdTreeSceneManager::GetHierarchyInterface()
	{
		return mHierarchyInterface;
	}

	SceneNode* KdTreeSceneManager::createSceneNode(void)
	{
		SceneNode* sn = new KdTreeSceneNode(this);
		assert(mSceneNodes.find(sn->getName()) == mSceneNodes.end());
		mSceneNodes[sn->getName()] = sn;
		return sn;
	}

	SceneNode* KdTreeSceneManager::createSceneNode(const String& name)
	{
		// Check name not used
		if (mSceneNodes.find(name) != mSceneNodes.end())
		{
			OGRE_EXCEPT(
				Exception::ERR_DUPLICATE_ITEM,
				"A scene node with the name " + name + " already exists",
				"KdTreeSceneManager::createSceneNode" );
		}

		SceneNode* sn = new KdTreeSceneNode(this, name);
		mSceneNodes[sn->getName()] = sn;
		return sn;
	}

	Entity * KdTreeSceneManager::createEntity(const String& entityName, const String& meshName)
	{
		Entity *ent = SceneManager::createEntity(entityName, meshName);

		for (int i = 0; i < (int)ent->getNumSubEntities(); ++i)
		{
			ent->getSubEntity(i)->setId(mCurrentEntityId);
		}

		// increase counter of entity id values
		++ mCurrentEntityId;

		return ent;
	}

	// make sure it's called only for non-empty nodes .. avoids one uneccessary funciton call
	void KdTreeSceneManager::_updateNode(KdTreeSceneNode *node)
	{
		//LogManager::getSingleton().logMessage("### _updateNode called for " + node->getName());

		/* Rebuild kdtree when it was wiped out
		 * Usually this happens only before the first frame
		 * The initial AABB shall enclose all objects present
		 * in the scene at the time of construction
		 * TODO: find a more appropriate place for it,
		 * e.g., before the first render call
		 * teh stupid thing is I can't find any other place so far ...
		 */
		if (!mKdTree)
		{
			mKdTree = new KdTree(mMaxDepth);
			mKdTree->setBuildMethod(mBuildMethod);
			mKdTree->build(static_cast<KdTreeSceneNode *>(mSceneRoot));
		}

		// if the node is in the tree and _updateNode was called, then the node was moved/rotated/resized/wahtever
		if (node->isAttached())
		{
			// TEST: perfomance when adding & removing for every frame
			//mKdTree->remove(node);
			//mKdTree->insert(node);
		}
		// there's a new node in town ... 
		else
		{			
			// inserting single nodes yields sub-optimal results
			// rebuilding tree takes too long
			// "What now?", spoke Zeus ...

			//mKdTree->insert(node);
			
			//delete mKdTree;
			//mKdTree = new KdTree(mMaxDepth);
			//mKdTree->build(static_cast<KdTreeSceneNode *>(mSceneRoot));
		}

	}

	void KdTreeSceneManager::_findVisibleObjects(Camera *cam, bool onlyShadowCasters)
	{
		getRenderQueue()->clear();
		if (mKdTree)
			mKdTree->queueVisibleObjects(cam, getRenderQueue(), onlyShadowCasters, mRenderMethod, mShowBoxes);
		//SceneManager::_findVisibleObjects(cam, onlyShadowCasters);
	}

	// TODO: looks too easy, verify if it works 
	void KdTreeSceneManager::_renderNodes(const KdRenderableList& nodelist, Camera * cam, 
		bool onlyShadowCasters, int leavePassesInQueue)
	{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		_deleteRenderedQueueGroups(leavePassesInQueue); //try avoiding modified ogre code
#else
		getRenderQueue()->clear();
#endif
		KdRenderableList::const_iterator it = nodelist.begin();
		KdRenderableList::const_iterator end = nodelist.end();
		while (it != end)
		{
			(*it)->queueObjects(cam, getRenderQueue(), onlyShadowCasters);
			it++;
		}

		SceneManager::_renderVisibleObjects();

	}

	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	const String KdTreeSceneManagerFactory::FACTORY_TYPE_NAME = "KdTreeSceneManager";
	//-----------------------------------------------------------------------
	void KdTreeSceneManagerFactory::initMetaData(void) const
	{
		mMetaData.typeName = FACTORY_TYPE_NAME;
		mMetaData.description = "Scene manager that organises the scene based on a kd-tree";
		mMetaData.sceneTypeMask = 0xFFFF; // support all types of scenes (hopefully)
		mMetaData.worldGeometrySupported = false;
	}
	//-----------------------------------------------------------------------
	SceneManager* KdTreeSceneManagerFactory::createInstance(
		const String& instanceName)
	{
		return new KdTreeSceneManager(instanceName, visManager);
	}
	//-----------------------------------------------------------------------
	void KdTreeSceneManagerFactory::destroyInstance(SceneManager* instance)
	{
		delete instance;
	}

} // namespace Ogre

